import yadagestep
import logging
from handlers.predicate_handlers import handlers as pred_handlers
from packtivity.statecontexts import load_provider

log = logging.getLogger(__name__)

class StageBase(object):
    '''
    Base class for workflow stages (i.e. extension rules)
    provides common datastructures and the extension predicate.
    Implementations are required to provide a schedule()
    method that is called upon apply
    '''

    def __init__(self, name, state_provider, dependencies=None):
        self.view = None
        self.name = name
        self.state_provider = state_provider
        self.depspec = dependencies

    def applicable(self, flowview):
        if not self.depspec:
            return True
        predicate = pred_handlers[self.depspec['dependency_type']]
        return predicate(flowview, self.depspec)

    def apply(self, flowview):
        self.view = flowview
        self.schedule()

    def schedule(self):
        raise NotImplementedError()

    def addStep(self, step):
        dependencies = [self.view.dag.getNode(k.stepid) for k in step.inputs]
        for d in dependencies:
            try:
                step.context.add_dependency(d.task.context)
            except AttributeError:
                pass
        return self.view.addStep(step, stage = self.name, depends_on=dependencies)

    def addWorkflow(self, rules, initstep):
        self.view.addWorkflow(rules, initstep=initstep, stage=self.name)

    #(de-)serialization
    def json(self):
        return {
            'name': self.name,
            'state_provider': self.state_provider.json() if self.state_provider else None,
            'dependencies': self.depspec
        }


class initStage(StageBase):
    '''
    simple stage that just adds a initializer step to the DAG
    '''

    def __init__(self, step, state_provider, dependencies):
        super(initStage, self).__init__('init', state_provider, dependencies)
        self.step = step

    def schedule(self):
        log.debug('initializing a scope with init step: %s',
                  self.step.prepublished)
        self.addStep(self.step)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        instance = cls(
            step = yadagestep.initstep.fromJSON(data['step']),
            state_provider = load_provider(data['state_provider']),
            dependencies = data['dependencies']
        )
        return instance

    def json(self):
        data = super(initStage, self).json()
        data.update(type='initStage', info='', step=self.step.json())
        return data


class jsonStage(StageBase):
    '''
    A stage that is defined via the JSON scheduler schemas
    '''

    def __init__(self, json, state_provider):
        self.stageinfo = json['scheduler']
        super(jsonStage, self).__init__(
            json['name'], state_provider, json['dependencies'])

    def __repr__(self):
        return '<jsonStage: {}>'.format(self.name)

    def schedule(self):
        #imported here to avoid circular dependency
        from handlers.scheduler_handlers import handlers as sched_handlers
        scheduler = sched_handlers[self.stageinfo['scheduler_type']]
        scheduler(self, self.stageinfo)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        return cls(
            json={
                'scheduler': data['info'],
                'name': data['name'],
                'dependencies': data['dependencies']
            },
            state_provider=load_provider(data['state_provider'])
        )

    def json(self):
        data = super(jsonStage, self).json()
        data.update(type='jsonStage', info=self.stageinfo)
        return data



import logging
import utils
import itertools
import copy

from expression_handlers import handlers as exprhandlers
from yadage.yadagestep import yadagestep, initstep, outputReference
from ..stages import jsonStage
from yadage.helpers import leaf_iterator

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()

# A scheduler does the following things:
#   - attached new nodes to the DAG
# - for each added step
#     - the step is given a name
#     - the step attributes are determined using the scheduler spec and context
#     - a list of used inputs (in the form of [stepname,outputkey,index])


def select_parameter(wflowview, parameter):
    '''
    Evaluates parameter expressions (if needed) in the context of a workflow view

    :param wflowview: the workflow view on which to evaluete possible value expressions
    :param parameter: either a non-dict value or a JSON-like dict for a
                      supported value expression
    :return: the parameter value 
    '''
    if type(parameter) is not dict:
        value = parameter
    else:
        handler = exprhandlers[parameter['expression_type']]
        value = handler(wflowview, parameter)
    return value

def finalize_value(wflowview, step, value, state):
    '''
    finalize a value by recursively resolving references and
    contextualizing it for the passed state context

    :param wflowview: the workflow view against which to resolve upstream references
    :param step: the step for which to track usage of upstream references
    :param value: the parameter value. May be a output reference, or a JSON value type
    :param state: the state context used to contextualize parameter values
    :return: finalized parameter value
    '''
    if type(value) == outputReference:
        step.used_input(value)
        v = value.pointer.resolve(wflowview.dag.getNode(value.stepid).result)
        return finalize_value(wflowview, step, v, state)
    if state:
        return state.contextualize_data(value)
    else:
        return value


def finalize_input(wflowview, step, jsondata, state):
    '''
    evaluate final values of step parameters by either resolving a
    reference to a upstream output and contextualizing stateful
    parameters. Also tracks usage of upstream references for the step

    :param wflowview: the workflow view view against which to resolve any upstream references
    :param step: the step that for which to track usage of upstream references
    :param jsondata: the prospective step parameters
    :param state: the state context 

    :return: finalized step parameters
    '''
    result = copy.deepcopy(jsondata)
    for leaf_pointer, leaf_value in leaf_iterator(jsondata):
        leaf_pointer.set(result,finalize_value(wflowview, step, leaf_value, state))
    return result

def step_or_init(name, spec, state_provider):
    '''
    create a named yadagestep of sub-workflow initstep object based on stage spec

    :param name: name of the eventual (init-)step
    :param spec: the stage spec
    :param state_provider: the stage's state provider

    :return: yadage or init step object
    '''
    if 'step' in spec:
        step_state = state_provider.new_state(name)
        return yadagestep(name=name, spec=spec['step'], context=step_state)
    elif 'workflow' in spec:
        return initstep('init {}'.format(name))

def addStepOrWorkflow(name, stage, step, spec):
    '''
    adds a step or a sub-workflow belonging to a stage this stage init step to the current workflow view
    
    :param str name: the name of the step or sub-workflow
    :param stage: the stage from which to use state context and workflow view
    :param step: either a yadagestep (for normal workflow steps) initstep object (for sub-workflows)
    :param spec: the stage spec

    :return: None
    '''
    if type(step) == initstep:
        new_provider = stage.state_provider.new_provider(name)
        subrules = [jsonStage(yml, new_provider) for yml in spec['workflow']['stages']]
        stage.addWorkflow(subrules, initstep=step)
    else:
        stage.addStep(step)

def get_parameters(spec):
    '''
    retrieve parameters from the spec

    :param spec: the stage spec
    :return: a JSON-like object of stage parameters
    '''
    return {x['key']: x['value']for x in spec['parameters']}

@scheduler('singlestep-stage')
def singlestep_stage(stage, spec):
    '''
    a simple state that adds a single step/workflow. The node is attached
    to the DAG based on used upstream outputs
    
    :param stage: common stage parent object 
    :param spec: stage JSON-like spec
    
    :return: None
    '''
    log.debug('scheduling singlestep stage with spec:\n%s', spec)

    step = step_or_init(name=stage.name, spec=spec, state_provider=stage.state_provider)
    ctx = step.context if hasattr(step, 'context') else stage.state_provider

    parameters = {
        k: select_parameter(stage.view, v) for k, v in get_parameters(spec).iteritems()
    }
    finalized = finalize_input(stage.view, step, parameters, ctx)
    addStepOrWorkflow(stage.name, stage, step.s(**finalized), spec)

def scatter(parameters, scatter):
    '''
    convert a parameter set and scatter definition into a list
    of single parameter sets.
    :param parameters: the parameter definition
    :param scatter: scattering method. One of 'zip' or 'cartesian'
    :return: list of parameter sets
    '''
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter['parameters']:
        to_scatter[scatpar] = commonpars.pop(scatpar)

    singlesteppars = []
    if scatter['method'] == 'zip':
        keys, zippable = zip(
            *[(k, v) for i, (k, v) in enumerate(to_scatter.iteritems())])

        for zipped in zip(*zippable):
            individualpars = dict(zip(keys, zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter['method'] == 'cartesian':
        for what in itertools.product(*[to_scatter[k] for k in scatter['parameters']]):
            individualpars = dict(zip(scatter['parameters'], what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars


@scheduler('multistep-stage')
def multistep_stage(stage, spec):
    '''
    a stage that attaches an array of nodes to the DAG. The number of nodes
    is determined by a scattering recipe. Currently two algs are supported

    - ``zip``: one or more arrays of length n are iterated through in lock-step.
       n nodes are added to the DAG where the  parameters values are set to
           the values in the iteration
    - ``cartesian``:  a cartesian product of a number of arrays (possibly different sizes)
       adds n1 x n2 x ... nj nodes.

    Nodes are attached to the DAG based on used upstream inputs

    :param stage: common stage parent object 
    :param spec: stage JSON-like spec
    
    :return: None
    '''
    log.debug('scheduling multistep stage with spec:\n%s', spec)
    parameters = {
        k: select_parameter(stage.view, v) for k, v in get_parameters(spec).iteritems()
    }
    singlesteppars = scatter(parameters, spec['scatter'])
    for i, pars in enumerate(singlesteppars):
        singlename = '{}_{}'.format(stage.name, i)
        step = step_or_init(name=singlename, spec=spec, state_provider = stage.state_provider)
        ctx = step.context if hasattr(step, 'context') else None
        finalized = finalize_input(stage.view, step, pars, ctx)
        addStepOrWorkflow(singlename, stage, step.s(**finalized), spec)



#!/usr/bin/env python
import os
import click
import jsonschema
import logging
import workflow_loader
import yadageschemas
import json
from helpers import WithJsonRefEncoder


logging.basicConfig(level=logging.ERROR)
log = logging.getLogger(__name__)


@click.command()
@click.argument('workflow')
@click.argument('toplevel', default='')
@click.argument('schemadir', default='')
@click.option('--stdout', '-s', default=False, is_flag=True)
def main(workflow, toplevel, schemadir, stdout):
    rc = 1
    if not toplevel:
        toplevel = os.getcwd()
    if not schemadir:
        schemadir = yadageschemas.schemadir
    try:
        data = workflow_loader.workflow(
            workflow, toplevel=toplevel, schemadir=schemadir, validate=True)
        if stdout:
            print json.dumps(data, cls=WithJsonRefEncoder)
        else:
            click.secho('workflow validates against schema', fg='green')
        rc = 0
    except jsonschema.exceptions.ValidationError:
        click.secho('workflow does not validate against schema', fg='red')
        raise
    except:
        log.exception('')
        click.secho(
            'this is not even wrong (non-ValidationError exception)', fg='red')

    if rc == 1:
        exc = click.exceptions.ClickException(
            click.style("validation failed", fg='red'))
        exc.exit_code = rc
        raise exc

if __name__ == '__main__':
    main()



import packtivity
import logging
import jsonpointer
from packtivity.statecontexts import load_state

log = logging.getLogger(__name__)

class outputReference(object):

    def __init__(self, stepid, pointer):
        self.stepid = stepid
        self.pointer = pointer

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        return cls(data['stepid'], jsonpointer.JsonPointer(data['pointer_path']))

    def json(self):
        return {
            'stepid': self.stepid,
            'pointer_path': self.pointer.path
        }

class StepBase(object):

    def __init__(self, name):
        self.name = name
        self.inputs = []
        self.attributes = {}
        self.prepublished = None

    def used_input(self, reference):
        self.inputs += [reference]

    #(de-)serialization
    def json(self):
        return {
            'name': self.name,
            'attributes': self.attributes,
            'prepublished': self.prepublished,
            'inputs': [x.json() for x in self.inputs]
        }

class initstep(StepBase):

    def __init__(self, name, initdata=None):
        super(initstep, self).__init__(name)
        self.prepublished = None
        if initdata is not None:
            self.s(**initdata)

    def s(self, **attributes):
        self.attributes = attributes
        self.prepublished = self.attributes
        return self

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        instance = cls(data['name'])
        instance.attributes = data['attributes']
        instance.prepublished = data['prepublished']
        instance.inputs = map(outputReference.fromJSON, data['inputs'])
        return instance

    def json(self):
        data = super(initstep, self).json()
        data.update(type='initstep')
        return data


class yadagestep(StepBase):

    def __init__(self, name, spec, context):
        super(yadagestep, self).__init__(name)
        self.spec = spec
        self.context = context

    def s(self, **attributes):
        self.attributes.update(**attributes)
        # attempt to prepublish output data merely from inputs
        # will still be None if not possible
        self.prepublished = packtivity.prepublish_default(self.spec, self.attributes, self.context)
        log.debug('parameters for yadagestep set to %s. prepublished result, if any: %s',
                    self.attributes,
                    self.prepublished
                  )
        return self

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        instance = cls(data['name'], data['spec'], load_state(data['context']))
        instance.attributes = data['attributes']
        instance.prepublished = data['prepublished']
        instance.inputs = map(outputReference.fromJSON, data['inputs'])
        return instance

    def json(self):
        data = super(yadagestep, self).json()
        data.update(
            type='yadagestep',
            spec=self.spec,
            context=self.context.json(),
        )
        return data



import json
import contextlib
import logging
import functools

import yadage.reset
import adage.controllerutils as ctrlutils
from adage.wflowcontroller import BaseController, InMemoryController
from yadage.wflow import YadageWorkflow


log = logging.getLogger(__name__)

class VariableProxy():
    @staticmethod
    def fromJSON(data):
        import packtivity.asyncbackends
        import yadage.backends.packtivitybackend
        if data['proxyname']=='InitProxy':
            return yadage.backends.packtivitybackend.InitProxy.fromJSON(data)
        elif data['proxyname']=='CeleryProxy':
            return packtivity.asyncbackends.CeleryProxy.fromJSON(data)
        else:
            raise RuntimeError('only celery support for now... found proxy with name: {}'.format(data['proxyname']))

def load_state_custom_deserializer(jsondata, backendstring = None):
    from clihelpers import setupbackend_fromstring
    if backendstring:
        backend = setupbackend_fromstring(backendstring)

    workflow = YadageWorkflow.fromJSON(
        jsondata,
        VariableProxy,
        backend
    )
    return workflow

def create_model_fromstring(modelidstring):
    modeltype, modelid = modelidstring.split(':')
    if modeltype == 'filebacked':
        return FileBackedModel(
            filename = modelid,
            deserializer = functools.partial(load_state_custom_deserializer, backendstring = 'celery'),
        )
    if modeltype == 'mongo':
        return MongoBackedModel(
            deserializer = functools.partial(load_state_custom_deserializer, backendstring = 'celery'),
            wflowid = modelid
        )

def setup_controller_fromstring(workflowobj, ctrlstring = 'inmem'):
    if ctrlstring == 'inmem':
        return InMemoryController(workflowobj, backend = None)
    elif ctrlstring.startswith('filebacked'):
        filename = ctrlstring.split(':')[-1]
        model   = FileBackedModel(
            filename = filename,
            deserializer = functools.partial(load_state_custom_deserializer, backendstring = 'celery'),
            initdata = workflowobj
        )
        return PersistentController(model)
    elif ctrlstring == 'mongo':
        model = MongoBackedModel(
            deserializer = functools.partial(load_state_custom_deserializer, backendstring = 'celery'),
            initdata = workflowobj
        )
        return PersistentController(model)
    else:
        raise RuntimeError('unknown workflow controller %s', ctrlstring)

class MongoBackedModel(object):
    def __init__(self, deserializer, connect_string = 'mongodb://localhost:27017/', initdata = None, wflowid = None):
        from pymongo import MongoClient
        from bson.objectid import ObjectId
        self.deserializer = deserializer
        self.client = MongoClient(connect_string)
        self.db = self.client.wflowdb
        self.collection = self.db.workflows
        if initdata:
            insertion = self.collection.insert_one(initdata.json())
            self.wflowid = insertion.inserted_id
            log.info('created new workflow object with id %s', str(self.wflowid))
        if wflowid:
            self.wflowid = ObjectId(wflowid)

    def commit(self, data):
        self.collection.replace_one({'_id':self.wflowid}, data.json())

    def load(self):
        return self.deserializer(self.collection.find_one({'_id':self.wflowid}))

class FileBackedModel(object):
    def __init__(self, filename, deserializer, initdata = None):
        self.filename = filename
        self.deserializer = deserializer
        if initdata:
            self.commit(initdata)

    def commit(self, data):
        '''
        :param data: data to commit to disk. needs to have '.json()' method
        commits data (possibly to persistent storage)
        '''
        with open(self.filename,'w') as statefile:
            json.dump(data.json(), statefile)

    def load(self):
        '''
        :return: the adage workflow object holding rules and the graph
        '''
        with open(self.filename) as statefile:
            return self.deserializer(json.load(statefile))

@contextlib.contextmanager
def model_transaction(self):
    '''
    param: model: a model object with .load() and .commit(data) methods
    '''
    self._adageobj = self.model.load()
    yield

    isvalid = self.validate()
    if not isvalid:
        #raise RuntimeError('was about to commit invalid data!')
        log.warning('commit is in valid %s', isvalid)
    self.model.commit(self._adageobj)

class PersistentController(BaseController):
    '''
    workflow controller, that explicltly calls transaction methods on non read-only operations on the workflow state
    '''
    def __init__(self, model, backend = None):
        super(PersistentController, self).__init__(backend)
        self.model = model
        self._adageobj = self.model.load()


    def transaction(self):
        return model_transaction(self)

    @property
    def adageobj(self):
        return self._adageobj

    def submit_nodes(self, nodeids):
        # log.info('about to submit')
        with self.transaction():
            nodes = [self._adageobj.dag.getNode(nodeid) for nodeid in nodeids]
            # log.info('submitting nodes to backend: %s', nodes)
            super(PersistentController,self).submit_nodes(nodes)

    def apply_rules(self, ruleids):
        with self.transaction():
            rules = [r for r in self._adageobj.rules if r.identifier in ruleids]
            super(PersistentController,self).apply_rules(rules)

    def sync_backend(self):
        with self.transaction():
            super(PersistentController,self).sync_backend()

    def applicable_rules(self):
        '''
        :return: return a list of rules whose predicate is fulfilled
        '''
        a = [x.identifier for x in ctrlutils.applicable_rules(self.adageobj)]
        return a

    def submittable_nodes(self):
        '''
        :return: a list of nodes with sucessfull and completed upstream
        '''
        return [x.identifier for x in ctrlutils.submittable_nodes(self.adageobj)]

    def reset_nodes(self, nodeids):
        '''
        :param nodes: 
        '''
        with self.transaction():
            yadage.reset.reset_steps(self._adageobj,nodeids)



import adage
import adage.backends
import os
import json
import workflow_loader
import clihelpers
import visualize
import serialize
import yadageschemas
import shutil
import logging
from packtivity.statecontexts.posixfs_context import LocalFSProvider, LocalFSState

from yadage.controllers import setup_controller_fromstring
from yadage.wflow import YadageWorkflow

log = logging.getLogger(__name__)

class YadageSteering():
    def __init__(self,loggername = __name__):
        self.log = logging.getLogger(loggername)
        self.workdir = None
        self.yadagedir = None
        self.controller = None
        self.rootprovider = None
        self.adage_kwargs = {}

    @property
    def workflow(self):
        return self.controller.adageobj

    def prepare_workdir(self, workdir, accept_existing_workdir = False, contextinit = None):
        self.workdir = workdir


        writable_context  = LocalFSState([workdir])
        self.rootprovider = LocalFSProvider(contextinit,writable_context)

        self.yadagedir = '{}/_yadage/'.format(workdir)

        if os.path.exists(self.yadagedir):
            if not accept_existing_workdir:
                raise RuntimeError('yadage meta directory exists. explicitly accept')
            self.log.info('yadage meta directory exists.. will remove and remake')
            shutil.rmtree(self.yadagedir)
        os.makedirs(self.yadagedir)
    
    def init_workflow(self, workflow, toplevel, initdata, ctrlsetup = 'inmem', initdir = None, search_initdir = True, validate = True, schemadir = yadageschemas.schemadir):
        ##check input data
        if search_initdir and initdir:
            clihelpers.discover_initfiles(initdata,os.path.realpath(initdir))

        workflow_json = workflow_loader.workflow(
            workflow,
            toplevel=toplevel,
            schemadir=schemadir,
            validate=validate
        )
        with open('{}/yadage_template.json'.format(self.yadagedir), 'w') as f:
            json.dump(workflow_json, f)
        workflowobj = YadageWorkflow.createFromJSON(workflow_json, self.rootprovider)
        if initdata:
            log.info('initializing workflow with %s',initdata)
            workflowobj.view().init(initdata)
        else:
            log.info('no initialization data')

        self.controller = setup_controller_fromstring(workflowobj, ctrlsetup)

    def adage_argument(self,**kwargs):
        self.adage_kwargs.update(**kwargs)

    def run_adage(self, backend, **adage_kwargs):
        self.controller.backend = backend
        self.adage_argument(**adage_kwargs)
        adage.rundag(controller = self.controller, **self.adage_kwargs)

    def serialize(self):
        serialize.snapshot(
            self.workflow,
            '{}/yadage_snapshot_workflow.json'.format(self.yadagedir),
            '{}/yadage_snapshot_backend.json'.format(self.yadagedir)
        )

    def visualize(self):
        visualize.write_prov_graph(self.yadagedir, self.workflow, vizformat='png')
        visualize.write_prov_graph(self.yadagedir, self.workflow, vizformat='pdf')

